# signature { license=0, seed=457995617, key="34f71b3782c1c27d95d762c9693e60ff" }
local now = os.clock
local cgi = require("cgi")
local mime = require("mime")
local http = require("socket.http")
local lfs = require("lfs")
local lower = string.lower

local runningApps={}
local maxUploadSize = 400*1024

-- from url.lua and PiL 20.3
local function unescape(s)
	s = string.gsub(s, "+", " ")
    return string.gsub(s, "%%(%x%x)", function(hex)
        return string.char(tonumber(hex, 16))
    end)
end

--
-- i18n = i(nternationalizatio)n 
-- function to find a localised text, if possible
--
local i18nTexts = {}
do
	local code, error = loadfile( "\\apps\\html\\i18n.txt" , "t", i18nTexts )
	if code then
		code()
		i18nTexts = i18nTexts.i18n
	end
end

function i18n( key, default)
	return i18nTexts[key] or default or key
end

---
-- list of known permissions (and texts to print them)
---
permText =
{
	OS_MEMORY   = i18n("use up to %d bytes of main memory");
	OS_KILL     = i18n("stop other applications");
	LFS_MKDIR   = i18n("create directories on disk");
	LFS_RMDIR   = i18n("delete directories on disk");
	LFS_SETATTR = i18n("change a file's attributes");
	INST_LIB    = i18n("install a library module");
	UPD_APPMGR  = i18n("update the application manager");
}
-- add default handler for unknown permission codes
setmetatable( permText, {__index = function( _,p ) return string.format(i18n("use system operation '%s'"), p) end} )

--
-- function(s) to "rm -rf" an application's directory
--
local attr = lfs.attributes
local function rm_rf(path)
	require ("permissions")
    for file in lfs.dir(path..'\\*') do
        if file ~= "." and file ~= ".." then
            local f = path..'\\'..file
            if attr( f, "directory") then
            	rm_rf(f)
            else
                os.remove(f)
            end
        end
    end
	lfs.rmdir(path)
end

local function saveTo( data, path )
	local start = now()

	os.remove(path)
	local f = io.open( path, "w")
	if f then
		f:write( data)
		f:close()
io.write( string.format("%s (%d bytes) saved in %0.2f s\n", path, #data, now()-start+0.001 ))
		return true
	end
end

--
-- make list of running applets
-- count booting and starting as well
--
local function findRunningApps()
	runningApps = {}
	for i=1,10 do
		local state,path = os.psinfo(i)
		if state and state>0 and state<4 then
			local name= path:lower():match( "\\([^\\]+)\\%1[-.]lua$")
			if name then
				runningApps[name]=i
				runningApps[i]=name
			end
		end
	end
end

--
-- function to pick from the temporary upload file manifest and icon.png
-- create the app's directory, and html if needed
-- and copy the <name>.pkg, the manifest.pkg and the icon
--
local function preInstall(posted)
	require ("permissions")
	local zip = require("zip")

	local tmp = "C:\\apps\\html\\manager.tmp"
	local mnf = "C:\\apps\\html\\manifest.tmp"
	saveTo( posted,      tmp, "w")
	local zipF = zip.open( tmp )
	if not zipF then return nil, i18n("package has wrong file format!") end

	local m = zipF:extract( "manifest" )
	if not m or not m.data then return nil, i18n("package contains no manifest") end
	if not saveTo( m.data, mnf ) then return nil, i18n("can't write manifest") end
	if os.license(mnf) == nil then return nil, i18n("manifest has an invalid signature") end

	local manifest = {}
	local code = loadfile( mnf, "t", manifest)
	if not code then return nil, i18n("manifest is not a valid Lua script") end
	code()
	if not manifest.name then return nil, i18n("manifest contains no valid name") end

	-- create directory if not present
	local path = "C:\\apps\\"..manifest.name
	lfs.mkdir( path )
	if lfs.attributes( path, "directory" ) then

		-- extract icon and save, if not already existing..
		local icon = zipF:extract( "html/icon.png")
		if icon and icon.data and not lfs.attributes( path.."\\html\\icon.png", "size" ) then
			lfs.mkdir( path.."\\html")
			saveTo (icon.data, path.."\\html\\icon.png")
		end
	end
	zipF:close()
	-- move received file and extracted manifest
	os.rename( tmp, path.."\\"..manifest.name..".pkg" )
	os.rename( mnf, path.."\\manifest.pkg" )
	return manifest.name
end

local function unInstall( appname )
	require ("permissions")
	local path= "C:\\apps\\"..appname
	if appname=="lib" or appname=="html" 
	or lfs.attributes(path.."\\"..appname..".lua", "size") then
		-- installation was an update, remove update packages only
		os.remove( path.."\\manifest.pkg")
		os.remove( path.."\\"..appname..".pkg")
	else
		-- installation was never complete: remove all
		rm_rf(path)
	end
end

local function doInstall( appname )
	require ("permissions")
	local zip = require("zip")

	local apppath = "C:\\apps\\"..appname.."\\"
	local pkg = apppath..appname..".pkg"
	local zipF = zip.open( pkg )
	if not zipF then return end

	local function unpackFiles( path, T, target )
		-- first the folders
		for name,subT in pairs( T ) do
			if type(subT) == "table" then
				lfs.mkdir( target..name )
				unpackFiles( path..name.."/", subT, target..name.."\\")
			end
		end

		-- then the files
		for _,name in ipairs( T ) do
			local fname = path..name
			local m = zipF:extract( fname )
			if m and m.data then saveTo( m.data, target..name) end
		end
	end

	local mfname = apppath.."manifest.pkg"
	mfname = mfname:gsub("/","\\")
	local manifest = {}
	local code = loadfile( mfname, "t", manifest)
	if code then
		code()
		local T = manifest.files
		unpackFiles( "", T, apppath)
	end
	zipF:close()
	os.remove( pkg )
	os.remove( mfname )
	os.remove( apppath..appname.."-lua" ) -- delete locked app
	local installer = apppath.."install.lua"
	if os.license( installer ) then pcall( dofile, installer ) end
end

local function getManifest( p )
	local M = {}
	local c = loadfile( p, "t", M)
	if c then
		c()
		return M
	end
end

local function getAppDetails( f )
	local p = "C:\\apps\\"..f

	if f:sub(1,1) =="."  or not lfs.attributes(p,"directory") then return end

    local app = {}

	app.name = f
	
	if lfs.attributes( p.."\\html\\icon.png", "size") then app.icon = true end
    local s= io.open(p.."\\"..f..".lua", "r")
    if s then
		app.canRun = true
		s:close()
    else
    	s= io.open(p.."\\"..f.."-lua", "r")
		if s then 
			app.isLocked = true
			app.canRun = true
			s:close()
		end
    end
    
	if lfs.attributes( p.."\\html\\config.cgi", "size") then app.config = true end
	if lfs.attributes( p.."\\html\\"..f..".cgi", "size") then app.home = f..".cgi" end
	if runningApps[ f ] then app.canStop = true	end

	app.M = getManifest( p.."\\manifest")
	app.N = getManifest( p.."\\manifest.pkg")

	app.license = (app.N or app.M or {license=3500}).license or 0;
	
	if app.N then return app end
	if not app.M then return end
	if f == "lib"  then return end
	if f == "html" then return end

	return app
end

local function paragraph( title, formatstring, ... )
	local t = select( 1, ...)
	if not t or t =="" then return "" end
	t = string.format( formatstring, ...)
	cgi.printf([[<h5>%s</h5>%s<br />]], title, t or "")
end

local function listPermissions( P )
	local ph=false;
	for k,v in pairs(P) do
		if not ph then
			cgi.print("<h5>",i18n("Required Permissons"),"</h5><br />",i18n("This application requires your permission to"),"<ul>")
			ph = true;
		end
		cgi.print("<li>");
		cgi.printf( permText[ k], v )
		cgi.print("</li/>");
	end
	if ph then cgi.print("</ul>") end
end

local function listRequirements( R )
	local function get( name )
		local loaded = package.loaded
		package.loaded={}
		local ok, L = pcall( require, name )
		package.loaded = loaded
		if ok then return L end
	end
	
	if R.licenses then
		local ph=false;
		for k,v in pairs(R.licenses) do
			local n = os.license(tonumber(k)) 
			local V= tonumber(v) or 1
			if n < V then
				if not ph then
					cgi.print("<h5>",i18n("Required Licenses"),"</h5><br />",i18n("This application requires to install"),"<ul>")
					ph = true;
				end
				cgi.printf("<li>%d instance(s) of License %d</li/>", V-n, k )
			end
		end
		if ph then cgi.print("</ul>") end
	end

	if R.libraries then
		local ph=false;
		for k,v in pairs(R.libraries) do
			local V=v
			local lib = get(k)
			local versionOK
			if not lib then 
				versionOK = false
				V=""
			else
				local imajor,iminor = (lib._VERSION or "version 0.0"):match("^%w+%s+(%d+)[.]?(%d*)")
				versionOK = type( v)~="string"
				imajor = tonumber(imajor or 0)
				iminor = tonumber(iminor or 0)
				if versionOK then 
					V = "0.0"
				else
					local wmajor,wminor = v:match("(%d+)[.]?(%d*)")
					wmajor = tonumber(wmajor or 0)
					wminor = tonumber(wminor or 0)
					versionOK =  imajor > wmajor or
								imajor == wmajor and
								iminor >= wminor 
				end
			end
			if not versionOK then
				if not ph then
					cgi.print("<h5>",i18n("Required Libraries"),"</h5><br />",i18n("This application requires to install"),"<ul>")
					ph = true;
				end
				cgi.printf("<li>%s %s</li/>", k, V)
			end
			lib = nil
			collectgarbage("collect")
		end
		if ph then cgi.print("</ul>") end
	end
end

local function geti18n()
	local set ={}

	local file = io.open( "\\apps\\html\\manager.cgi" )
	local text = file:read("*a")
	file:close()
	
	for _, key in text:gmatch( "i18n%s*[(]%s*([\"\'])(.-)%1" ) do 
		set[key] = key
	end

	cgi.header( "Content-Disposition", "attachment; filename=i18n.txt" );
	
	for row in io.lines( "\\apps\\html\\i18n.txt" ) do
		local line = row:gsub('\r','')
		local key = line:match('%["([^"]*)"%]')
		if key then 
			if set[key] then
				set[key] = nil
			else
				line = line .. "-\45 unused !"
			end
		else
			if line:find("^}") then -- end of array, add new entries
				local sorted = {}
				for key in pairs( set ) do table.insert( sorted, key ) end

				cgi.printf("-\45\r\n-\45 new texts\r\n-\45\r\n")
				table.sort( sorted )
				for _, key in ipairs( sorted ) do
					cgi.printf( '["%s"]=\r\n[[%s]];\r\n\r\n' , key, key:gsub('\\n','\r\n') )
				end
			end
		end
		cgi.printf("%s\r\n", line)
	end
	return true
end

--
-- check authorisation
--
local authorised = false

if not authorised then
	local s = CGI.AUTHORIZATION
	if s then
		local _,port
		local xmlpro = require("xmlpro.core")
		if xmlpro then 
			port = xmlpro.port
		else
			_,port = cgi.socket:getsockname() 
		end

		s=string.match(s, "^%s*Basic%s+(.*)" )
		s=mime.unb64(s)
		local user,pw = string.match( s, "^([^:]+):(.+)")

		-- test if credentials have installer or MF level
		XmlProUrl = string.format( "http://%s:%s@%s:%d/dbms.xml", user, pw, "127.0.0.1", port)
		local x = http.request( XmlProUrl, "<get><mb mb=\"9002\"><k k=\"2\">[1-2]</k></mb></get>" )
		if x:find("i=") then
			authorised = true
		else
			require("socket").sleep(3)
		end
	end
end

if not authorised then
	cgi.status = "401 Unauthorized"
	cgi.header("WWW-Authenticate","Basic realm=\"PBX Applet Management\"")
	cgi.print("Unauthorised!")
	return
end

findRunningApps()
local cmd,value = "", ""
if CGI.REQUEST_METHOD == "POST" then
	local name, posted, err
	local length = 1*(CGI.CONTENT_LENGTH or 0)
	if length > maxUploadSize then 
		MsgBox = i18n("Package is too large").."\\n"
		-- read the content into the /dev/nil...
		while length>8192 do
			cgi.receive(8192)
			length = length -8192
		end
		cgi.receive(length)
		length = 0
	end
	if length >0 then
		posted = cgi.receive(length)
		local enctype, boundary = string.match( (CGI.CONTENT_TYPE or ""), "%s*([^;]+);%s*boundary=(.*)$" )
		if enctype == "multipart/form-data" then
			boundary = string.gsub( boundary, "[][().%+-*?^$]", "%%%1")
			posted = string.match( posted, boundary..".-\r\n\r\n(.-)\r\n%-%-"..boundary)
				name,err = preInstall(posted)
				if name then 
					value = name
				else
					print ("error:", err) 
					MsgBox = i18n("Package upload failed").."\\n"..err
				end
		else
			posted = unescape( posted )
			cmd, value = string.match( posted, "cmd=([^=]+)=(.+)")
		end
	end
elseif CGI.QUERY_STRING >"" then
	cmd, value = string.match( CGI.QUERY_STRING, "&?([^=]+)=([^&]+)")
end
value = lower(value or "")
local hasLic3500 = os.license(3500)>0

--
-- check command
--
local commands = {
	run = function( value )
		if not runningApps[ value ] then 
			local name = string.format("C:\\apps\\%s\\%s", value, value)
			local f = io.open(name..".lua")
			if f then
				f:close()
				name = name ..".lua"
			else
				f = io.open(name.."-lua")
				if f then
					f:close()
					name = name .."-lua"
				else 
					return
				end
			end
							
			os.execute( name)
			runningApps[ value ] = -1 -- mark as started
		end
		return true;
	end;

	stop = function( value )
		require ("permissions")
		local m = runningApps[ value ]
		if m then
			os.kill(runningApps[ value ])
			findRunningApps()
			runningApps[ value ] = nil -- it will be stopped soon
		end
	end;

	remove = function( value )
		if value~="lib" and value~="html"  then -- these are protected
			rm_rf("C:\\apps\\"..value);
		end
	end;

	cancel = unInstall;
	install = doInstall;
	lock   = function( value ) os.rename( "C:\\apps\\"..value.."\\"..value..".lua", value.."-lua");	return true; end;
	unlock = function( value ) os.rename( "C:\\apps\\"..value.."\\"..value.."-lua", value..".lua"); return true; end;
	no3500 = function() hasLic3500 = false; print("Developer License 3500 ignored") end;
	geti18n = geti18n;
}

local f = commands[ cmd ]
if f and  f( value ) then 
	cgi.header( "Content-Type",'text/plain; charset="UTF-8"');
	-- cgi.print("ok")
	return
end

local apps = {}
for name in lfs.dir("C:\\apps\\*") do
	local app = getAppDetails( lower(name) )
	if app then
		table.insert( apps, app)
		apps[app.name] = app
	end
end

local selected = { }
if (#apps>0) then
	if value=="" or not apps[value] then value = apps[1].name end
	selected[ value ] = [[class="sel"]]
end

--
-- generate output
--
cgi.header( "Content-Type", [[text/html; charset="UTF-8"]] )
cgi.header( "X-UA-Compatible", "IE=edge" ) -- override IE's "show intranet sites in compatibility mode"
cgi.printf( [[
<!DOCTYPE HTML>
<html>
<head>
<meta charset="UTF-8" />
<title>%s</title>
<link rel="SHORTCUT ICON" href="/html/apps/favicon.ico">
<link rel="stylesheet" type="text/css" href="common.css" >
<link rel="stylesheet" type="text/css" href="manager.css" >
<script>var MsgBox="%s";
var confirm1="%s";
var confirm2="%s";
]], 
i18n("NEC PBX Application Manager"),
MsgBox or "",
i18n("Do you really want to remove this application\nand all its files and directories ?"):gsub('\n', '\\n'),
i18n("Are you sure ?")
)

cgi.print( [[var AppNames=[ ]] )
local comma=""
for _,app in ipairs(apps) do
	cgi.printf([[%s"%s"]], comma, app.name:lower())
	comma=','
end
cgi.print( [[ ];</script>]] )


cgi.printf( [[
<script src="manager.js" type="text/javascript" ></script>
</head>
<body>
	<h1>%s <small>%s</small></h1>
	<h2>
		<form name="upload" action="manager.cgi" method="post" enctype="multipart/form-data">
      <a class="button"><div class="upload">
        <input type="file" id="uploadfile" name="uploadfile"/>%s</div>
      </a>
		</form>
%s</h2>
	<div class="apps">
		<form name="post" action="manager.cgi" method="post"><input type="hidden" name="cmd" /></form>
		<ul class="apps" id="applist">
]],i18n("Application Manager"), "1.2.4", i18n("Install"), i18n("Installed Applications"))

--
-- list of tabs on the left
--
for _,app in ipairs(apps) do
	local M = app.N or app.M
	local c= M.description
	cgi.printf([[<li  style="background-image:url(/html/apps%s/icon.png);" onclick="details('%s')" id="list-%s" %s><h3>%s&nbsp;%s</h3>%s</li>]],
			(app.icon and "/"..app.name or ""),
			app.name,    -- onclick()
			app.name,    -- id
			selected[lower(app.name)] or first,
			c.name   or "Applet", -- H4 title
			M.version or "",
			c.summary or ""
	)
	first=""
end

cgi.print([[</ul></div>
<div class="details" id="appdetails">]])
--
-- stack of detail pages on the right
--

for _,app in ipairs(apps) do
	local M = app.N or app.M
	local c= M.description

	if app.license == 0 or hasLic3500 then
		licensed = 1 -- always on
	else
		licensed = os.license( app.license )
		if licensed < 1 then
			app.canRun = false
			app.config = false
		end
	end

	cgi.printf( [[<div id="detail-%s" %s><h4 class="buttons">%s ]],
		app.name,
		selected[lower(app.name)] or first,
		c.name   or i18n("Details") -- H4 title
	)
	
	if app.name~="html" and app.name~="lib" then
		if app.config then
			cgi.printf([[<a class="button" id="%s-config" href="/html/apps/%s/config.cgi" target="_blank"><img src="config.png" />%s</a>]], app.name, app.name, i18n("Configure"))
		end

		if app.canRun then
			cgi.printf([[<a class="button hide" id="%s-run" href="javascript:doRun('%s')"><img src="play.png" />%s</a>]], app.name, app.name, i18n("Start"))
		end

		cgi.printf([[<a class="button hide"  id="%s-stop"   href="javascript:doStop('%s')"><img src="stop.png" />%s</a>]], app.name, app.name, i18n("Stop"))
		cgi.printf([[<a class="button hide"  id="%s-remove"   href="javascript:doRemove('%s')"><img src="trash.png" />%s</a>]], app.name, app.name, i18n("Remove"))

		if app.canRun and licensed >0 then
			cgi.printf([[<a class="button%s" id="%s-unlock" href="javascript:doUnlock('%s')"><img src="unlock.png" />%s</a>]], app.isLocked and "" or " hide", app.name, app.name, i18n("Enable"))
			cgi.printf([[<a class="button%s" id="%s-lock"   href="javascript:doLock('%s')"><img src="lock.png" />%s</a>]],    app.isLocked and " hide" or "", app.name, app.name,i18n("Disable"))
		end

		for _,button in ipairs(c.buttons or {}) do
			if button.img and button.href then
				cgi.printf([[<a class="button" href="%s"><img src="%s" />%s</a>]], button.href, button.img, button[1] or "")
			end 
		end
	end
	
	cgi.print("</h4>")

	paragraph( i18n("Summary:"), "<br />%s", c.summary)
	paragraph( i18n("Description:"), "<br />%s", c.detailed)
	paragraph( i18n("Home Page:"), [[<a href="%s">%s</a>]], c.homepage, c.homepage)
	paragraph( i18n("Maintained by:"), "%s", c.maintainer)
	paragraph( i18n("Update URL:"), [[<a href="%s">%s</a>]], c.updateURL, c.updateURL)
	paragraph( i18n("Required License Code:"), "%s", (app.license==0) and i18n("free") or tostring(app.license) )
	if licensed <= 0 then
		cgi.printf("<em>%s<br />%s</em><br />",	i18n("This license is not installed."), i18n("It is required to start or configure the application"))
	end
	paragraph( i18n("Copyright:"), "%s", c.copyright)

	if app.N then
		cgi.printf( [[<h4 class="buttons hr">%s
<a class="button" id="%s-install" href="javascript:doInstall('%s')"><img src="ok.png" />%s</a>
<a class="button" id="%s-cancel" href="javascript:doCancel( '%s')"><img src="cancel.png" />%s </a></h4>]],
			i18n("Uploaded Package ready for Installation"),
			app.name,	app.name, i18n("Install"),
			app.name,	app.name, i18n("Cancel")
		)
		if app.N.permissions then listPermissions( app.N.permissions ) end
		if app.N.requirements then listRequirements( app.N.requirements ) end
		if app.N.version then paragraph( i18n("New Version:"), "%s", app.N.version) end
	end
	if app.M then
		paragraph( i18n("Installed Version:"), "%s", app.M.version or "")
	end
	first = ""
	cgi.print("</div>")
end
cgi.print("</div>")
cgi.print([[</body></html>]])

-- history
-- 2014-04-02 changed all path separators to backslash, new NucleusOS is a little bit more picky
-- 2014-04-04 starting machines are considered now as running
--            a script can only be started if it is not already running (or started)
--            after stopping one script, all processes are re-scanned
-- 2014-07-07 modified to present an updated package in apps\html or apps\lib
--            added automatic start of apps\...\install.lua on install
-- 2015-08-31 added i18n() functionality.
-- 2015-09-04 changed handling of manifest.pkg file. Now a package MUST list 
--            "manifest" in the file list to get it copied to its folder
